// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2002 Martin Schweiger
//                   All rights reserved
//
// Atlantis.cpp
// Reference implementation of Atlantis (Space Shuttle) vessel
// class module
//
// RMS, grappling and MMU capabilities by Robert Conley
// ==============================================================

#include "Atlantis.h"
#include <stdio.h>
#include <fstream.h>

bool reset_mmu = false;
OBJHANDLE hMMU;
double jettison_time = 0.0;
bool reset_sat = false;
OBJHANDLE hSAT;

// Calculate lift coefficient [Cl] as a function of aoa (angle of attack) over -Pi ... Pi
// Implemented here as a piecewise linear function
double LiftCoeff (double aoa)
{
	const int nlift = 9;
	static const double AOA[nlift] = {-180*RAD,-60*RAD,-30*RAD,-1*RAD,15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nlift]  = {       0,      0,   -0.2,     0,   0.8,     1,   0.8,     0,      0};
	static const double SCL[nlift] = {(CL[1]-CL[0])/(AOA[1]-AOA[0]), (CL[2]-CL[1])/(AOA[2]-AOA[1]),
		                              (CL[3]-CL[2])/(AOA[3]-AOA[2]), (CL[4]-CL[3])/(AOA[4]-AOA[3]),
									  (CL[5]-CL[4])/(AOA[5]-AOA[4]), (CL[6]-CL[5])/(AOA[6]-AOA[5]),
									  (CL[7]-CL[6])/(AOA[7]-AOA[6]), (CL[8]-CL[7])/(AOA[8]-AOA[7])};
	for (int i = 0; i < nlift-1 && AOA[i+1] < aoa; i++);
	return CL[i] + (aoa-AOA[i])*SCL[i];
}

extern void GetSRB_State (double met, double &thrust_level, double &prop_level);

// ==============================================================
// Specialised vessel class Atlantis
// ==============================================================

// Constructor
Atlantis::Atlantis (OBJHANDLE hObj, int fmodel)
: VESSEL(hObj, fmodel)
{
	status          = 0;
	gear_status     = DOOR_CLOSED;
	gear_proc       = 0.0;
	gear_moving     = false;
	gear_endanim    = false;
	door_status     = DOOR_CLOSED;
	door_proc       = 0.0;
	bManualSeparate = false;
	vis             = NULL;

	// propellant resources
	ph_oms          = NULL;
	ph_tank         = NULL;
	ph_srb          = NULL;
	thg_main        = NULL;
	thg_srb         = NULL;

	// preload meshes
	hOrbiterMesh = oapiLoadMeshGlobal ("Atlantis");
	hTankMesh    = oapiLoadMeshGlobal ("Atlantis_tank");
	hSRBMesh     = oapiLoadMeshGlobal ("Atlantis_srb");


	DefineAnimations();
	center_arm      = false;
	arm_moving      = false;
	arm_endanim     = false;
	bManualSeparate = false;
	sat_jettisoned  = true;
	do_switch        = false;
	sat_ofs         = _V(0,0,0);
	ofs_sts_sat     = _V(0,0,0);      
	do_eva          = false;
	do_plat         = false;
	do_cargostatic  = false;
	vis             = NULL;
	arm_angle       = RAD*10;
	arm_max         = 0.0;
	strcpy(sat_name,"sts_sat");
	strcpy(sat_mesh_name,"sts_sat");
	varm_proc          =0;
	arm_proc           =0;
	arm_action         =ARM_SHOULDER_PITCH;
	arm_shoulder_pitch = 0.0;
	arm_shoulder_yaw   = 0.0;
	arm_elbow_pitch    = 0.0;
	arm_wrist_pitch    = 0.0;
	arm_wrist_yaw      = 0.0;
	arm_wrist_roll     = 0.0;
	sat_grapple        =_V(0,0,0);//_V(0,2.34065,.181733);
	sat_grapple_norm   =_V(0,1,0);
	cargo_static_ofs   =_V(0,0,0);
	sat_grappled       = false;
}

// Read status from scenario file
void Atlantis::LoadState (FILEHANDLE scn, void *vs)
{
    char *line;
	double met = 0.0; // mission elapsed time
	double srbtime = 0.0;
	double sts_sat_x = 0.0;
	double sts_sat_y = 0.0;
	double sts_sat_z = 0.0;
	double grapple_x = 0.0;
	double grapple_y = 0.0;
	double grapple_z = 0.0;
	int jettisoned=1;
	int grappled=0;

	while (oapiReadScenario_nextline (scn, line)) {
        if (!strnicmp (line, "CONFIGURATION", 13)) {
            sscanf (line+13, "%d", &status);
		} else if (!strnicmp (line, "MET", 3)) {
			sscanf (line+3, "%lf", &met);
		} else if (!strnicmp (line, "CARGODOOR", 9)) {
			sscanf (line+9, "%d%lf", &door_status, &door_proc);
		} else if (!strnicmp (line, "GEAR", 4)) {
			sscanf (line+4, "%d%lf", &gear_status, &gear_proc);
		} else if (!strnicmp (line, "SRB_IGNITION_TIME", 17)) {
			sscanf (line+17, "%lf", &srbtime);
 		} else if (!strnicmp (line, "SAT_JETTISONED", 14)) {
			sscanf (line+14, "%d", &jettisoned);
		} else if (!strnicmp (line, "SAT_OFS_X", 9)) {
			sscanf (line+9, "%lf", &sts_sat_x);
		} else if (!strnicmp (line, "SAT_OFS_Y", 9)) {
			sscanf (line+9, "%lf", &sts_sat_y);
		} else if (!strnicmp (line, "SAT_OFS_Z", 9)) {
			sscanf (line+9, "%lf", &sts_sat_z);
		} else if (!strnicmp (line, "SAT_NAME", 8)) {
			sscanf (line+8, "%s", sat_name);
		} else if (!strnicmp (line, "SAT_MESH", 8)) {
			sscanf (line+8, "%s", sat_mesh_name);
		} else if (!strnicmp (line, "CARGO_STATIC_MESH", 17)) {
			sscanf (line+17, "%s", cargo_static_mesh_name);
			do_cargostatic = true;
		} else if (!strnicmp (line, "CARGO_STATIC_OFS", 16)) {
			sscanf (line+16, "%lf%lf%lf", &cargo_static_ofs.x, &cargo_static_ofs.y, &cargo_static_ofs.z);
		} else if (!strnicmp (line, "ARM_SH_P", 8)) {
			sscanf (line+8, "%lf", &arm_shoulder_pitch);
		} else if (!strnicmp (line, "ARM_SH_Y", 8)) {
			sscanf (line+8, "%lf", &arm_shoulder_yaw);
		} else if (!strnicmp (line, "ARM_EL_P", 8)) {
			sscanf (line+8, "%lf", &arm_elbow_pitch);
		} else if (!strnicmp (line, "ARM_WR_P", 8)) {
			sscanf (line+8, "%lf", &arm_wrist_pitch);
		} else if (!strnicmp (line, "ARM_WR_Y", 8)) {
			sscanf (line+8, "%lf", &arm_wrist_yaw);
		} else if (!strnicmp (line, "ARM_WR_R", 8)) {
			sscanf (line+8, "%lf", &arm_wrist_roll);
		} else if (!strnicmp (line, "SAT_GRAPPLE", 11)) {
			sscanf (line+11, "%lf%lf%lf", &grapple_x, &grapple_y, &grapple_z);
 		} else if (!strnicmp (line, "SAT_GRAPPLED", 12)) {
			sscanf (line+12, "%d", &grappled);
        } else {
            ParseScenarioLineEx (line, vs);
			// unrecognised option - pass to Orbiter's generic parser
        }
    }
	ofs_sts_sat.x=sts_sat_x;
	ofs_sts_sat.y=sts_sat_y;
	ofs_sts_sat.z=sts_sat_z;

	sat_grapple.x = ofs_sts_sat.x + grapple_x;
	sat_grapple.y = ofs_sts_sat.y + grapple_y;
	sat_grapple.z = ofs_sts_sat.z + grapple_z;

	if (jettisoned==1)
		sat_jettisoned=true;
	else
		sat_jettisoned=false;

	if (grappled==1)
		sat_grappled=true;
	else
		sat_grappled=false;

	switch (status) {
	case 0:
		SetLaunchConfiguration();
		break;
	case 1:
		SetPostLaunchConfiguration (met);
		break;
	case 2:
		SetOrbiterTankConfiguration();
		break;
	case 3:
		SetOrbiterConfiguration();
		break;
	}

	// update animation states
	SetAnimState (anim_door, door_proc);

	// launch animations if required
	if (gear_status == DOOR_OPENING || gear_status == DOOR_CLOSING)
		OperateLandingGear();
}

// Write status to scenario file
void Atlantis::SaveState (FILEHANDLE scn)
{
	char cbuf[256];
	int jettisoned=0;
	VECTOR3 grapple_pt;

	// default vessel parameters
	SaveDefaultState (scn);

	// custom parameters
	oapiWriteScenario_int (scn, "CONFIGURATION", status);

	if (status == 1)
		oapiWriteScenario_float (scn, "MET", oapiGetSimTime()-t0);

	sprintf (cbuf, "%d %0.4f", door_status, door_proc);
	oapiWriteScenario_string (scn, "CARGODOOR", cbuf);

	sprintf (cbuf, "%d %0.4f", gear_status, gear_proc);
	oapiWriteScenario_string (scn, "GEAR", cbuf);
	oapiWriteScenario_float (scn, "SAT_OFS_X", ofs_sts_sat.x);
	oapiWriteScenario_float (scn, "SAT_OFS_Y", ofs_sts_sat.y);
	oapiWriteScenario_float (scn, "SAT_OFS_Z", ofs_sts_sat.z);

	oapiWriteScenario_float (scn, "ARM_SH_P", arm_shoulder_pitch);
	oapiWriteScenario_float (scn, "ARM_SH_Y", arm_shoulder_yaw);
	oapiWriteScenario_float (scn, "ARM_EL_P", arm_elbow_pitch);
	oapiWriteScenario_float (scn, "ARM_WR_P", arm_wrist_pitch);
	oapiWriteScenario_float (scn, "ARM_WR_Y", arm_wrist_yaw);
	oapiWriteScenario_float (scn, "ARM_WR_R", arm_wrist_roll);

	if (sat_jettisoned)
		jettisoned=1;
	else
		jettisoned=0;
	oapiWriteScenario_int (scn, "SAT_JETTISONED", jettisoned);

	oapiWriteScenario_string (scn, "SAT_NAME", sat_name);
	oapiWriteScenario_string (scn, "SAT_MESH", sat_mesh_name);

	grapple_pt = sat_grapple - ofs_sts_sat;
	oapiWriteScenario_vec    (scn, "SAT_GRAPPLE", grapple_pt);
	if (sat_grappled)
		jettisoned=1;
	else
		jettisoned=0;
	oapiWriteScenario_int (scn, "SAT_GRAPPLED", jettisoned);

	if (do_cargostatic) {
		oapiWriteScenario_string (scn, "CARGO_STATIC_MESH", cargo_static_mesh_name);
		if (cargo_static_ofs.x || cargo_static_ofs.y || cargo_static_ofs.z)
			oapiWriteScenario_vec (scn, "CARGO_STATIC_OFS", cargo_static_ofs);
	}
}

// launch configuration setup (orbiter + tank + SRB)
void Atlantis::SetLaunchConfiguration (void)
{
	int i;

	// *********************** physical parameters *********************************

	SetSize (30.0);
	SetEmptyMass (ORBITER_EMPTY_MASS + TANK_EMPTY_MASS + 2*SRB_EMPTY_MASS);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(687.4,849.5,189.4));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(179.1,176.8,29.3));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_LAUNCH_ORBITER, _V(0,1,0), _V(0,0,-1));
	SetTouchdownPoints (_V(0,-10,-53.3), _V(-7,7,-53.3), _V(7,7,-53.3));

	// ************************* propellant specs **********************************

	if (!ph_oms)  ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	if (!ph_tank) ph_tank = CreatePropellantResource (TANK_MAX_PROPELLANT_MASS);    // main tank
	if (!ph_srb)  ph_srb  = CreatePropellantResource (SRB_MAX_PROPELLANT_MASS*2.0); // SRB's
	SetDefaultPropellantResource (ph_tank); // display main tank level in generic HUD

	// *********************** thruster definitions ********************************

	// The main and SRB thrusters are defined so as to minimise the angular momentum
	// when engaged at maximum thrust. Dynamic gimbaling is not required until the SRBs
	// start to lose power since the centre of gravity is assumed static.
	// However the resulting linear force vector has a component in +y ("up") direction

	ClearThrusterDefinitions();

	// orbiter main thrusters
	th_main[0] = CreateThruster (OFS_LAUNCH_ORBITER+_V(-1.6,-0.2,-16.0), _V( 0.04994,0.0,0.99875), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_main[1] = CreateThruster (OFS_LAUNCH_ORBITER+_V( 1.6,-0.2,-16.0), _V(-0.04994,0.0,0.99875), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_main[2] = CreateThruster (OFS_LAUNCH_ORBITER+_V( 0.0, 2.7,-16.0), _V( 0.0,-0.13,1), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	thg_main = CreateThrusterGroup (th_main, 3, THGROUP_MAIN);
	for (i = 0; i < 3; i++) AddExhaust (th_main[i], 8.0, 1.0);

	// SRBs
	th_srb[0] = CreateThruster (OFS_LAUNCH_RIGHTSRB+_V(0.0,0.0,-21.8), _V(0,0.023643,0.999720), SRB_THRUST, ph_srb, SRB_ISP0, SRB_ISP1);
	th_srb[1] = CreateThruster (OFS_LAUNCH_LEFTSRB +_V(0.0,0.0,-21.8), _V(0,0.023643,0.999720), SRB_THRUST, ph_srb, SRB_ISP0, SRB_ISP1);
	thg_srb = CreateThrusterGroup (th_srb, 2, THGROUP_USER);
	SURFHANDLE tex = oapiRegisterExhaustTexture ("Exhaust2");
	for (i = 0; i < 2; i++) AddExhaust (th_srb[i], 16.0, 2.0, tex);

	// attitude - this is temporary
	// attitude adjustment during launch phase should really be done via SRB gimbaling
	CreateAttControls_Launch();

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_LAUNCH_ORBITER, !sat_jettisoned);
	AddTankVisual    (OFS_LAUNCH_TANK);
	AddSRBVisual     (OFS_LAUNCH_RIGHTSRB);
	AddSRBVisual     (OFS_LAUNCH_LEFTSRB);

	status = 0;
}

void Atlantis::SetPostLaunchConfiguration (double met)
{
	SetLaunchConfiguration();
	t0 = -met; // reference time (liftoff)
	status = 1;
}

// Configuration after booster separation (orbiter + tank)
void Atlantis::SetOrbiterTankConfiguration (void)
{
	int i;
	VECTOR3 ofs;
	
	// *********************** physical parameters *********************************

	SetSize (28.8);
	SetEmptyMass (ORBITER_EMPTY_MASS + TANK_EMPTY_MASS);
	SetCW (0.2, 0.5, 1.5, 1.5);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(646.1,603.0,141.5));
	SetRotDrag (_V(0.7,0.1,0.3));
	SetPMI (_V(173.3,161.0,24.0));
	SetTrimScale (0.05);
	SetLiftCoeffFunc (0); // simplification: we assume no lift during launch phase
	SetDockParams (ORBITER_DOCKPOS+OFS_WITHTANK_ORBITER, _V(0,1,0), _V(0,0,-1));
	SetTouchdownPoints (_V(0,-5,30), _V(-10,-10,-30), _V(10,0,-30));

	// ************************* propellant specs **********************************

	if (!ph_oms)  ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	if (!ph_tank) ph_tank = CreatePropellantResource (TANK_MAX_PROPELLANT_MASS);    // main tank
	SetDefaultPropellantResource (ph_tank); // display main tank level in generic HUD

	// *********************** thruster definitions ********************************

	// orbiter main engines
	ofs = OFS_WITHTANK_ORBITER;
	if (thg_main) { // main engines already defined - just modify parameters
		SetThrusterRef (th_main[0], ofs+_V(-1.6,-0.2,-16.0));
		SetThrusterDir (th_main[0], _V( 0.0624,-0.1789,0.9819));
		SetThrusterRef (th_main[1], ofs+_V( 1.6,-0.2,-16.0));
		SetThrusterDir (th_main[1], _V(-0.0624,-0.1789,0.9819));
		SetThrusterRef (th_main[2], ofs+_V( 0.0, 2.7,-16.0));
		SetThrusterDir (th_main[2], _V( 0.0,   -0.28516,0.9585));
	} else {        // create main engines
		th_main[0] = CreateThruster (ofs+_V(-1.6,-0.2,-16.0), _V( 0.0624,-0.1789,0.9819), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		th_main[1] = CreateThruster (ofs+_V( 1.6,-0.2,-16.0), _V(-0.0624,-0.1789,0.9819), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		th_main[2] = CreateThruster (ofs+_V( 0.0, 2.7,-16.0), _V( 0.0,   -0.28516,0.9585), ORBITER_MAIN_THRUST, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
		thg_main = CreateThrusterGroup (th_main, 3, THGROUP_MAIN);
		for (i = 0; i < 3; i++) AddExhaust (th_main[i], 8.0, 1.0);
	}
	if (!ThrusterGroupDefined (THGROUP_ATT_PITCHUP))
		CreateAttControls_Launch();

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_WITHTANK_ORBITER, !sat_jettisoned);
	AddTankVisual    (OFS_WITHTANK_TANK);

	status = 2;
}

// Configuration after tank separation (orbiter only)
void Atlantis::SetOrbiterConfiguration (void)
{
	int i;

	// *********************** physical parameters *********************************

	SetSize (19.6);
	SetEmptyMass (ORBITER_EMPTY_MASS);
	SetPMI (_V(78.2,82.1,10.7));
	SetCW (ORBITER_CW[0], ORBITER_CW[1], ORBITER_CW[2], ORBITER_CW[3]);
	SetCrossSections (ORBITER_CS);
	if (gear_status == DOOR_OPEN)
		SetTouchdownPoints (_V(0,-4,17), _V(-3.96,-5,-4.3), _V(3.96,-5,-4.3)); // gear wheel tips
	else
		SetTouchdownPoints (_V(0,-2.5,14), _V(-8,-2.8,-9), _V(8,-2.8,-9)); // belly landing
	SetLiftCoeffFunc (LiftCoeff);

	if (GetFlightModel() >= 1) { // realistic flight model
		SetPitchMomentScale (1e-5);
		SetBankMomentScale (2e-5);
		if (gear_status == DOOR_OPEN) {
			SetCW (ORBITER_CW[0]+ORBITER_CW_GEAR[0],
				   ORBITER_CW[1]+ORBITER_CW_GEAR[1],
				   ORBITER_CW[2]+ORBITER_CW_GEAR[2],
				   ORBITER_CW[3]+ORBITER_CW_GEAR[3]);
			SetCrossSections (ORBITER_CS+ORBITER_CS_GEAR);
		}
	}

	SetRotDrag (_V(0.5,1.0,1.0));

	SetDockParams (ORBITER_DOCKPOS, _V(0,1,0), _V(0,0,-1));
	// for docking attachment in payload bay

	// ************************* propellant specs **********************************

	if (!ph_oms) ph_oms  = CreatePropellantResource (ORBITER_MAX_PROPELLANT_MASS); // OMS propellant
	SetDefaultPropellantResource (ph_oms); // display OMS tank level in generic HUD

	// *********************** thruster definitions ********************************

	// OMS (Orbital Manouevering System)
	th_main[0] = CreateThruster (_V(-2.05,3.45,-14.2), _V( 0.1,-0.2349,0.9669), ORBITER_OMS_THRUST, ph_oms, ORBITER_OMS_ISP0, ORBITER_OMS_ISP1);
	th_main[1] = CreateThruster (_V( 2.05,3.45,-14.2), _V(-0.1,-0.2349,0.9669), ORBITER_OMS_THRUST, ph_oms, ORBITER_OMS_ISP0, ORBITER_OMS_ISP1);
	thg_main = CreateThrusterGroup (th_main, 2, THGROUP_MAIN);
	for (i = 0; i < 2; i++) AddExhaust (th_main[i], 4.0, 0.5);

	// RCS (Reaction Control System)
	CreateAttControls_RCS ();

	// ************************ visual parameters **********************************

	ClearMeshes();
	AddOrbiterVisual (OFS_ZERO, !sat_jettisoned);

	status = 3;
}

void Atlantis::CreateAttControls_Launch()
{
	// we should only require pitch and bank, but yaw is also defined to make KILLROT work
	// no linear attitude modes are defined at this stage
	THRUSTER_HANDLE th_rot[2];
	th_rot[0] = CreateThruster (_V(0,0, 25), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-25), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHUP);
	th_rot[0] = CreateThruster (_V(0,0, 25), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-25), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_PITCHDOWN);
	th_rot[0] = CreateThruster (_V( 8,0,0), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(-8,0,0), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKLEFT);
	th_rot[0] = CreateThruster (_V( 8,0,0), _V(0,-1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(-8,0,0), _V(0, 1,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_BANKRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V( 1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V(-1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWRIGHT);
	th_rot[0] = CreateThruster (_V(0,0, 10), _V(-1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	th_rot[1] = CreateThruster (_V(0,0,-10), _V( 1,0,0), MAX_ATT_LAUNCH, ph_tank, ORBITER_MAIN_ISP0, ORBITER_MAIN_ISP1);
	CreateThrusterGroup (th_rot, 2, THGROUP_ATT_YAWLEFT);
}

void Atlantis::CreateAttControls_RCS()
{
	// set of attitude thrusters (idealised). The arrangement is such that no angular
	// momentum is created in linear mode, and no linear momentum is created in rotational mode.
	THRUSTER_HANDLE th_att_rot[4], th_att_lin[4];
	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15.5), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-15.5), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 15.5), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15.5), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_PITCHUP);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_PITCHDOWN);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_UP);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_DOWN);
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V( 0,  -2.5 ,16  ), _V(0,-1,0));
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V(-0.4,-2.5 ,16  ), _V(0,-1,0));
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V( 0.4,-2.5, 16  ), _V(0,-1,0));
	AddExhaust (th_att_rot[1], 0.6, 0.078, _V(-3.4, 3.3,-14.4), _V(0, 1,0));
	AddExhaust (th_att_rot[1], 0.6, 0.078, _V( 3.4, 3.3,-14.4), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V( 0,   1.6 ,17  ), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V(-0.4, 1.55,17  ), _V(0, 1,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V( 0.4, 1.55,17  ), _V(0, 1,0));
	AddExhaust (th_att_rot[3], 0.6, 0.078, _V(-3.4, 1.6,-14.4), _V(0,-1,0));
	AddExhaust (th_att_rot[3], 0.6, 0.078, _V( 3.4, 1.6,-14.4), _V(0,-1,0));

	th_att_rot[0] = th_att_lin[0] = CreateThruster (_V(0,0, 15.5), _V(-1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = th_att_lin[3] = CreateThruster (_V(0,0,-15.5), _V( 1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = th_att_lin[2] = CreateThruster (_V(0,0, 15.5), _V( 1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = th_att_lin[1] = CreateThruster (_V(0,0,-15.5), _V(-1,0,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_YAWLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_YAWRIGHT);
	CreateThrusterGroup (th_att_lin,   2, THGROUP_ATT_LEFT);
	CreateThrusterGroup (th_att_lin+2, 2, THGROUP_ATT_RIGHT);
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V( 1.9,-0.6, 17.6), _V( 1,0,0));
	AddExhaust (th_att_rot[1], 0.6, 0.078, _V(-4,   2.5,-14.4), _V(-1,0,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V(-1.9,-0.6, 17.6), _V(-1,0,0));
	AddExhaust (th_att_rot[3], 0.6, 0.078, _V( 4,   2.5,-14.4), _V( 1,0,0));

	th_att_rot[0] = CreateThruster (_V( 2.7,0,0), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[1] = CreateThruster (_V(-2.7,0,0), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[2] = CreateThruster (_V(-2.7,0,0), _V(0, 1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_rot[3] = CreateThruster (_V( 2.7,0,0), _V(0,-1,0), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_rot,   2, THGROUP_ATT_BANKLEFT);
	CreateThrusterGroup (th_att_rot+2, 2, THGROUP_ATT_BANKRIGHT);
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V( 2.0, -0.6, 17.2), _V( 0.174,-0.985,0));
	AddExhaust (th_att_rot[0], 0.6, 0.078, _V( 3.4,  1.5,-14.0), _V( 0,    -1,    0));
	AddExhaust (th_att_rot[1], 0.6, 0.078, _V(-1.9, -0.6, 17.6), _V(-0.174, 0.985,0));
	AddExhaust (th_att_rot[1], 0.6, 0.078, _V(-3.4,  3.3,-14.0), _V( 0,     1,    0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V(-2.0, -0.6, 17.2), _V(-0.174,-0.985,0));
	AddExhaust (th_att_rot[2], 0.6, 0.078, _V(-3.4,  1.5,-14.0), _V( 0,    -1,    0));
	AddExhaust (th_att_rot[3], 0.6, 0.078, _V( 1.9, -0.6, 17.6), _V( 0.174, 0.985,0));
	AddExhaust (th_att_rot[3], 0.6, 0.078, _V( 3.4,  3.3,-14.0), _V( 0,     1,    0));

	th_att_lin[0] = CreateThruster (_V(0,0,-16), _V(0,0, 1), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	th_att_lin[1] = CreateThruster (_V(0,0, 16), _V(0,0,-1), ORBITER_RCS_THRUST, ph_oms, ORBITER_RCS_ISP0, ORBITER_RCS_ISP1);
	CreateThrusterGroup (th_att_lin,   1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_att_lin+1, 1, THGROUP_ATT_BACK);
	AddExhaust (th_att_lin[0], 0.6, 0.078, _V(-3.4,2.5,-15  ), _V(0,0,-1));
	AddExhaust (th_att_lin[0], 0.6, 0.078, _V( 3.4,2.5,-15  ), _V(0,0,-1));
	AddExhaust (th_att_lin[1], 0.6, 0.078, _V( 0,  1.2, 18.2), _V(0,0, 1));
	AddExhaust (th_att_lin[1], 0.6, 0.078, _V(-0.4,1.15,18.2), _V(0,0, 1));
	AddExhaust (th_att_lin[1], 0.6, 0.078, _V( 0.4,1.15,18.2), _V(0,0, 1));
}

// Define animation sequences for moving parts
void Atlantis::DefineAnimations (void)
{
	static UINT RCargoDoorGrp[3] = {90,92,93};
	static ANIMCOMP RCargoDoor = {
		RCargoDoorGrp, 3, 0.0, 0.6,
		2.8, 1.35, 0,         // rotation reference
		0,0,1,                // rotation axis
		(float)(-175.5*RAD),  // rotation angle
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT LCargoDoorGrp[3] = {91,94,95};
	static ANIMCOMP LCargoDoor = {
		LCargoDoorGrp, 3, 0.05, 0.65,
		-2.8, 1.35, 0,        // rotation reference
		0,0,1,                // rotation axis
		(float)(175.5*RAD),   // rotation angle
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT RRadiatorGrp[1] = {92};
	static ANIMCOMP RRadiator = {
		RRadiatorGrp, 1, 0.7, 1.0,
		2.8, 1.35, 0,
		0,0,1,
		(float)(35.5*RAD),
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};
	static UINT LRadiatorGrp[1] = {94};
	static ANIMCOMP LRadiator = {
		LRadiatorGrp, 1, 0.7, 1.0,
		-2.8, 1.35, 0,
		0, 0, 1,
		(float)(-35.5*RAD),
		0, 0,
		MESHGROUP_TRANSFORM::ROTATE
	};

	// Register animation for cargo doors
	anim_door = RegisterAnimSequence (0.0);
	AddAnimComp (anim_door, &RCargoDoor);
	AddAnimComp (anim_door, &LCargoDoor);
	AddAnimComp (anim_door, &RRadiator);
	AddAnimComp (anim_door, &LRadiator);
}

void Atlantis::AddOrbiterVisual (const VECTOR3 &ofs, bool add_sat)
{
	sat_ofs=ofs;
	sat_ofs.x += ofs_sts_sat.x;
	sat_ofs.y += ofs_sts_sat.y;
	sat_ofs.z += ofs_sts_sat.z;
	AddMesh (hOrbiterMesh, &ofs);

	if (do_cargostatic) {
		VECTOR3 cofs = ofs + cargo_static_ofs;
		AddMesh (cargo_static_mesh_name, &cofs);
	}

	if (do_plat)
	{
		VECTOR3 plat_ofs = _V(-2.59805, 1.69209, -5.15524);
		AddMesh("shuttle_eva_plat", &plat_ofs);
	}
	if (add_sat)
	{
		AddMesh (sat_mesh_name,&sat_ofs);
	}

	SetCameraOffset (_V(ofs.x,ofs.y+2.4,ofs.z+14));
}

void Atlantis::AddTankVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hTankMesh, &ofs);
	SetMeshVisibleInternal (meshidx, true);
}

void Atlantis::AddSRBVisual (const VECTOR3 &ofs)
{
	UINT meshidx;
	meshidx = AddMesh (hSRBMesh, &ofs);
	SetMeshVisibleInternal (meshidx, true);
	int id = AddExhaustRef (EXHAUST_CUSTOM, _V(ofs.x, ofs.y, ofs.z-21.8), 0, 0);
	if (ofs.x < 0.0) srb_id1 = id;
	else             srb_id2 = id;
}

void Atlantis::SeparateBoosters (double met)
{
	// Create SRB's as individual objects
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.flag[0] = 3;
	GetSRB_State (met, vs.eng_main, vs.fuel);
	Local2Rel (OFS_LAUNCH_RIGHTSRB, vs.rpos);
	vs.arot.z += 0.25*PI;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-SRB1");
	oapiCreateVessel (name, "Atlantis_SRB", vs);
	Local2Rel (OFS_LAUNCH_LEFTSRB, vs.rpos);
	vs.arot.z += 0.5*PI;
	name[strlen(name)-1] = '2';
	oapiCreateVessel (name, "Atlantis_SRB", vs);

	// Remove SRB's from Shuttle instance
	DelPropellantResource (ph_srb);
	DelThrusterGroup (thg_srb, THGROUP_USER, true);

	// reconfigure
	SetOrbiterTankConfiguration();
	ShiftCentreOfMass (OFS_LAUNCH_ORBITER-OFS_WITHTANK_ORBITER);
}

void Atlantis::SeparateTank (void)
{
	// Create Tank as individual object
	VESSELSTATUS vs;
	GetStatus (vs);
	vs.eng_main = vs.eng_hovr = 0.0;
	VECTOR3 ofs = OFS_WITHTANK_TANK;
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	VECTOR3 vel = {0,-1,0};
	Local2Rel (ofs, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;
	vs.vrot.x = -0.02;
	vs.status = 0;
	char name[256];
	strcpy (name, GetName()); strcat (name, "-Tank");
	oapiCreateVessel (name, "Atlantis_Tank", vs);

	// Remove Tank from shuttle instance
	DelPropellantResource (ph_tank);

	// main engines are done
	DelThrusterGroup (thg_main, THGROUP_MAIN, true);

	// clear launch attitude control system
	DelThrusterGroup (THGROUP_ATT_PITCHUP, true);
	DelThrusterGroup (THGROUP_ATT_PITCHDOWN, true);
	DelThrusterGroup (THGROUP_ATT_BANKLEFT, true);
	DelThrusterGroup (THGROUP_ATT_BANKRIGHT, true);
	DelThrusterGroup (THGROUP_ATT_YAWLEFT, true);
	DelThrusterGroup (THGROUP_ATT_YAWRIGHT, true);

	// reconfigure
	SetOrbiterConfiguration ();
	ShiftCentreOfMass (OFS_WITHTANK_ORBITER);
}


void Atlantis::SeparateSatellite (void)
{
	VESSELSTATUS vs;
	GetStatus (vs);
	char name[256];
	strcpy (name, GetName()); 
	strcat (name, "-Sat");
	vs.status = 0; // force "freeflight" status
	
	vs.eng_main = vs.eng_hovr = 0.0;
	VECTOR3 ofs = OFS_ZERO;
	VECTOR3 vel;
	if (sat_grappled) {
		ofs = ofs_sts_sat;
		SatellitePos (&ofs);  // get current satellite position
		// get arm rotation matrix
		VECTOR3 x0 = _V(1,0,0), y0 = _V(0,1,0), z0 = _V(0,0,1);
		SatelliteRot (&x0);
		SatelliteRot (&y0);
		SatelliteRot (&z0);
		// global rotation matrix of the parent
		MATRIX3 R;
		GetRotationMatrix (R);
		// combine both (we only calculate the elements required to build arot)
		double m11 = R.m11*x0.x + R.m12*x0.y + R.m13*x0.z;
		double m12 = R.m11*y0.x + R.m12*y0.y + R.m13*y0.z;
		double m13 = R.m11*z0.x + R.m12*z0.y + R.m13*z0.z;
		double m23 = R.m21*z0.x + R.m22*z0.y + R.m23*z0.z;
		double m33 = R.m31*z0.x + R.m32*z0.y + R.m33*z0.z;
		vs.arot.x = atan2 (m23, m33);
		vs.arot.y = -asin (m13);
		vs.arot.z = atan2 (m12, m11);
		GrappleNormal (&vel);
		vel *= 0.1;
		//vel = _V(0,0,0); // if releasing from grapple, don't give push
	} else {
		ofs.x += ofs_sts_sat.x;
		ofs.y += ofs_sts_sat.y;
		ofs.z += ofs_sts_sat.z;
		vel = _V(0,.2,0);
	}
	VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
	Local2Rel (ofs, vs.rpos);
	GlobalRot (vel, rofs);
	vs.rvel.x = rvel.x+rofs.x;
	vs.rvel.y = rvel.y+rofs.y;
	vs.rvel.z = rvel.z+rofs.z;

	hSAT = VESSEL::Create (name, sat_name, vs);
	jettison_time = oapiGetSimTime();
	// Remove Tank from shuttle instance
	SetOrbiterConfiguration ();
}

void Atlantis::ToggleGrapple (void)
{
	if (sat_grappled) {
		// should ungrapple be the same as jettison?
		// RSC No until we are able to retrieve satellites we should
		// just have this leave the mesh in place.
		//bManualSeparate = true;
		sat_grappled = false;
	} else {
		// remember arm position
		sat_shoulder_pitch = arm_shoulder_pitch;
		sat_shoulder_yaw = arm_shoulder_yaw;
		sat_elbow_pitch = arm_elbow_pitch;
		sat_wrist_pitch = arm_wrist_pitch;
		sat_wrist_yaw = arm_wrist_yaw;
		sat_wrist_roll = arm_wrist_roll;
		sat_grappled = true;
	}
}

void Atlantis::SeparateMMU (void)
{
	// Create Tank as individual object
	VESSELSTATUS vs;
	GetStatus (vs);
	char name[256];
	strcpy (name, GetName()); strcat (name, "-MMU");
	hMMU = VESSEL::Create (name, "Nasa_MMU", vs);
	jettison_time = oapiGetSimTime();
	reset_mmu = true;
	// Remove Tank from shuttle instance
	SetOrbiterConfiguration ();
}

void Atlantis::AutoMainGimbal ()
{
	// Implement automatic gimbal adjustment for main engines to correct for
	// SRB thrust variations. We adjust only the upper engine

	double        F_srb  = SRB_THRUST         *GetThrusterLevel (th_srb[0]);
	double        F_main = ORBITER_MAIN_THRUST*GetThrusterLevel (th_main[0]);
	if (F_main) {
		double        M_srb =   2.51951112*F_srb;  // angular moment from SRB thrust
		double        M_m2  = -12.02495   *F_main; // angular moment from the two lower main engines in neutral position
		double        M_0   = M_srb + M_m2;
		static double ry    =   8.92;  // upper main thruster: y-offset
		static double rz    = -23.795; // upper main thruster: z-offset
		static double ry2   = ry*ry;
		static double rz2   = rz*rz;
		double        F2    = F_main*F_main;
		double        M2    = M_0*M_0;
		double        term  = -M2 + F2*(ry2+rz2);
		double        arg1  = max (0.0, rz2 * term);
		double        scale = 1.0/(F_main*(ry2+rz2));
		double        dz    = (M_0*ry + sqrt (arg1))*scale;
		double		  arg2  = max (0.0, ry2 * term);
		double        dy    =-(M_0*rz + sqrt (arg2))*scale;
		if (!arg1 || !arg2) {
			double len = hypot(dy,dz);
			dy /= len, dz /= len;
		}
		SetThrusterDir (th_main[2], _V(0,dy,dz));
	}
}

void Atlantis::LaunchClamps ()
{
	VECTOR3 F, T, r = _V(0,0,0), Fc = _V(0,0,0), Tc = _V(0,0,0);
	GetThrusterMoment (th_srb[0], F, T);
	Fc.z = -2*F.z;
	Tc.x =  2*T.x;
	GetThrusterMoment (th_main[0], F, T);
	Fc.z -= 2*F.z;
	Fc.y -= 2*F.y;
	Tc.x += 2*T.x;
	GetThrusterMoment (th_main[2], F, T);
	Fc.z -= F.z;
	Fc.y -= F.y;
	Tc.x += T.x;
	r.z = (Fc.y ? Tc.x/Fc.y : 0);
	AddForce (Fc, r);
}

void Atlantis::Timestep (double simt)
{
	double met;
	int i;
	OBJHANDLE hvessel;

	switch (status) {
	case 0: // launch configuration
		if (GetEngineLevel (ENGINE_MAIN) > 0.95) {
			status = 1; // launch
			t0 = simt + SRB_STABILISATION_TIME;   // store designated liftoff time
		} else
			AutoMainGimbal();
		break;
	case 1: // SRB's ignited
		met = simt-t0;
		if (met > SRB_SEPARATION_TIME || bManualSeparate) { // separate boosters
			SeparateBoosters (met);
			bManualSeparate = false;
		} else {
			// extract current thrust level and propellant level as a function of time
			double thrust_level, prop_level;
			GetSRB_State (met, thrust_level, prop_level);
			for (i = 0; i < 2; i++)
				SetThrusterLevel (th_srb[i], thrust_level);
			if (met < 0.0) LaunchClamps ();
			AutoMainGimbal();
		}
		break;

	case 2: // post SRB separation
		if (GetPropellantMass (ph_tank) < 1.0 || bManualSeparate) {
			SeparateTank();
			bManualSeparate = false;
		}
		break;
	case 3:
		if (bManualSeparate && !sat_jettisoned)
		{
			SeparateSatellite();
			sat_jettisoned=true;
			SetOrbiterConfiguration();
			if (do_switch)
			{
				char name[256];
				strcpy (name, GetName()); strcat (name, "-Sat");
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			}
		}
		else if (bManualSeparate && sat_jettisoned)
		{
			if (do_switch)
			{
				char name[256];
				strcpy (name, GetName()); strcat (name, "-Sat");
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			}
		}
		if (do_eva)
		{
			char name[256];
			strcpy (name, GetName()); strcat (name, "-MMU");
			hvessel=oapiGetVesselByName(name);
			if (hvessel == 0)
			{
				SeparateMMU();
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			}
			else
			{
				hvessel=oapiGetVesselByName(name);
				if (hvessel != 0)
					oapiSetFocusObject(hvessel);
			};
		};
		do_eva=false;
		bManualSeparate=false;
		break;
	}

	VESSEL *aVessel;
	VESSELSTATUS vs;

#ifdef UNDEF
	if (reset_sat && simt-jettison_time > .01)
	{
		GetStatus(vs);
		vs.eng_main = vs.eng_hovr = 0.0;
		VECTOR3 ofs = OFS_ZERO;
		ofs.x += ofs_sts_sat.x;
		ofs.y += ofs_sts_sat.y;
		ofs.z += ofs_sts_sat.z;
		VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
		VECTOR3 vel = {0,2,0};
		Local2Rel (ofs, vs.rpos);
		GlobalRot (vel, rofs);
		vs.rvel.x = rvel.x+rofs.x;
		vs.rvel.y = rvel.y+rofs.y;
		vs.rvel.z = rvel.z+rofs.z;
		aVessel = new VESSEL (hSAT, 1);
		aVessel->DefSetState(&vs);
		reset_sat=false;
	}	
#endif

	if (reset_mmu && simt-jettison_time > .01)
	{
		GetStatus(vs);
		vs.eng_main = vs.eng_hovr = 0.0;
		VECTOR3 ofs = OFS_ZERO;
		ofs.x += OFS_MMU.x;
		ofs.y += OFS_MMU.y;
		ofs.z += OFS_MMU.z;
		VECTOR3 rofs, rvel = {vs.rvel.x, vs.rvel.y, vs.rvel.z};
		VECTOR3 vel = {0,0,0};
		Local2Rel (ofs, vs.rpos);
		GlobalRot (vel, rofs);
		vs.rvel.x = rvel.x+rofs.x;
		vs.rvel.y = rvel.y+rofs.y;
		vs.rvel.z = rvel.z+rofs.z;
		aVessel = new VESSEL (hMMU, 1);
		aVessel->DefSetState(&vs);
		reset_mmu=false;
	}

	if (door_status >= DOOR_CLOSING) { // animate cargo doors
		double da = oapiGetSimStep() * DOOR_OPERATING_SPEED;
		if (door_status == DOOR_CLOSING) {
			if (door_proc > 0.0) door_proc = max (0.0, door_proc-da);
			else                 door_status = DOOR_CLOSED;
		} else { // door opening
			if (door_proc < 1.0) door_proc = min (1.0, door_proc+da);
			else                 door_status = DOOR_OPEN;
		}
		SetAnimState (anim_door, door_proc);
	}

	if (center_arm)
	{
		if (arm_wrist_roll != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_WRIST_ROLL;
				arm_max   = -arm_wrist_roll;
				RevertArm();
			}
		}
		else if (arm_wrist_yaw != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_WRIST_YAW;
				arm_max   = -arm_wrist_yaw;
				RevertArm();
			}
		}
		else if (arm_wrist_pitch != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_WRIST_PITCH;
				arm_max   = -arm_wrist_pitch;
				RevertArm();
			}
		}
		else if (arm_elbow_pitch != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_ELBOW_PITCH;
				arm_max   = -arm_elbow_pitch;
				RevertArm();
			}
		}
		else if (arm_shoulder_yaw != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_SHOULDER_YAW;
				arm_max   = -arm_shoulder_yaw;
				RevertArm();
			}
		}
		else if (arm_shoulder_pitch != 0)
		{
			if (!arm_moving)
			{
				arm_action= ARM_SHOULDER_PITCH;
				arm_max   = -arm_shoulder_pitch;
				RevertArm();
			}
		}
		else
		{
			center_arm=false;
		}
	}
	
	if (arm_moving) 
	{ 
		double dt = oapiGetSimStep();
		double da = dt * ARM_OPERATING_SPEED;
		if (arm_max < 0) {
			if (arm_endanim) {
				arm_moving = arm_endanim = false;
				UnregisterAnimation();
			} else if ((arm_proc -= da) <= arm_max) {
				arm_proc    = arm_max;
				arm_endanim = true;
			}
		} else {
			if (arm_endanim) {
				arm_moving = arm_endanim = false;
				arm_max = 0;
				UnregisterAnimation();
			} else if ((arm_proc += da) >= arm_max) {
				arm_proc    = arm_max;
				arm_endanim = true;
			}
		}
	}

	if (gear_moving) { // update landing gear status
		double dt = oapiGetSimStep();
		double da = dt * GEAR_OPERATING_SPEED;
		double pgear_proc = gear_proc;
		if (gear_status == DOOR_CLOSING) {
			if (gear_endanim) {
				gear_status = DOOR_CLOSED;
				gear_moving = gear_endanim = false;
				UnregisterAnimation();
			} else if ((gear_proc -= da) <= 0.0) {
				gear_proc    = 0.0;
				gear_endanim = true;
				SetTouchdownPoints (_V(0,-2.5,14), _V(-8,-2.8,-9), _V(8,-2.8,-9));
				// gear retracted: set contact points for belly landing
			}
		} else if (gear_status = DOOR_OPENING) {
			if (gear_endanim) {
				gear_status = DOOR_OPEN;
				gear_moving = gear_endanim = false;
				UnregisterAnimation();
			} else if ((gear_proc += da) >= 1.0) {
				gear_proc    = 1.0;
				gear_endanim = true;
				SetTouchdownPoints (_V(0,-4,17), _V(-3.96,-5,-4.3), _V(3.96,-5,-4.3));
				// gear lowered: set gear contact points
			}
		}
		// update aerodynamics for current gear status
		if (GetFlightModel() >= 1) {
			double delta = gear_proc-pgear_proc;
			VECTOR3 cs;
			GetCrossSections (cs);
			SetCrossSections (cs + ORBITER_CS_GEAR*delta);
			double cw[4];
			GetCW (cw[0], cw[1], cw[2], cw[3]);
			for (int i = 0; i < 4; i++) cw[i] += ORBITER_CW_GEAR[i]*delta;
			SetCW (cw[0], cw[1], cw[2], cw[3]);
		}
	}
}

void Atlantis::Jettison ()
{
	switch (status) {
	case 0:
	case 3:               // nothing to do
		break;
	case 1:               // abandon boosters
		SeparateBoosters (oapiGetSimTime()-t0); 
		break;
	case 2:               // abandon tank
		SeparateTank();
		break;
	}
}

// Update moving parts of the orbiter's visual: payload bay doors and gear
// This should only be called when the visual exists, e.g. from within
// ovcVisualCreated or ovcAnimate

void Atlantis::UpdateMesh ()
{
	int grp_ofs;
	double delta;
	MESHGROUP_TRANSFORM mt;

	if (delta = arm_proc-varm_proc) { // need to bring pilot door up to date
		float da = (float)delta;
		this->ZeroArm();
		if (arm_action==ARM_SHOULDER_PITCH)
			arm_shoulder_pitch += da;
		else if (arm_action==ARM_SHOULDER_YAW)
			arm_shoulder_yaw   += da;
		else if (arm_action==ARM_ELBOW_PITCH)
			arm_elbow_pitch    += da;
		else if (arm_action==ARM_WRIST_PITCH)
			arm_wrist_pitch    += da;
		else if (arm_action==ARM_WRIST_YAW)
			arm_wrist_yaw      += da;
		else if (arm_action==ARM_WRIST_ROLL)
			arm_wrist_roll     += da;

		this->RotateArm();
		varm_proc = arm_proc;
	}

	if (gear_proc-vgear_proc) { // need to bring landing gear up to date
		mt.nmesh = 0;
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		grp_ofs = ORBITER_MESHGROUP_GEAR;

		// gear doors - note that the doors open during the first part of
		// the animation sequence (gear_proc < 0.5), while the gears are
		// lowered during the second part (gear_proc > 0.3)
		double alpha_tgt = (gear_proc  < 0.5 ?  gear_proc*1.1*PI : 0.55*PI);
		double alpha_now = (vgear_proc < 0.5 ? vgear_proc*1.1*PI : 0.55*PI);
		float da = (float)(alpha_tgt-alpha_now);
		if (da) {
			// left nose wheel door
			mt.P.rotparam.axis  = _V(0,0.195,0.981);
			mt.P.rotparam.angle = da;
			mt.P.rotparam.ref   = _V(-0.52,-2.19,17);
			for (mt.ngrp = grp_ofs; mt.ngrp < grp_ofs+2; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// right nose wheel door inside
			mt.P.rotparam.angle = -da;
			mt.P.rotparam.ref   = _V(0.52,-2.19,17);
			for (mt.ngrp = grp_ofs+2; mt.ngrp < grp_ofs+4; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// right gear door
			mt.P.rotparam.axis  = _V(0.0,0.0328,0.9994);
			mt.P.rotparam.angle = -da*0.95f;
			mt.P.rotparam.ref   = _V(4.8f,-2.47f,-4.69f);
			for (mt.ngrp = grp_ofs+6; mt.ngrp < grp_ofs+8; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// left gear door
			mt.P.rotparam.angle = da*0.95f;
			mt.P.rotparam.ref   = _V(-4.8f,-2.47f,-4.69f);
			for (mt.ngrp = grp_ofs+8; mt.ngrp < grp_ofs+10; mt.ngrp++)
				MeshgroupTransform (vis, mt);
		}

		// landing gear
		alpha_tgt = (gear_proc  > 0.3 ? ( gear_proc-0.3)*1.43*0.53*PI : 0.0);
		alpha_now = (vgear_proc > 0.3 ? (vgear_proc-0.3)*1.43*0.53*PI : 0.0);
		da = (float)(alpha_tgt-alpha_now);
		if (da) {
			// nose wheel
			mt.P.rotparam.angle = da;
			mt.P.rotparam.axis = _V(1,0,0);
			mt.P.rotparam.ref = _V(0,-1.5,16.95);
			for (mt.ngrp = grp_ofs+4; mt.ngrp < grp_ofs+6; mt.ngrp++)
				MeshgroupTransform (vis, mt);

			// left+right gear
			mt.P.rotparam.ref = _V(0,-2,-4.3);
			for (mt.ngrp = grp_ofs+10; mt.ngrp < grp_ofs+14; mt.ngrp++)
				MeshgroupTransform (vis, mt);
		}
		vgear_proc = gear_proc;
	}
}

void Atlantis::OperateLandingGear ()
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	if ((gear_status == DOOR_OPENING || gear_status == DOOR_CLOSING) && !gear_moving) {
		gear_moving = true;
		RegisterAnimation();
	}
}

void Atlantis::RevertLandingGear ()
{
	if (status < 3) return;
	// operate landing gear only once the orbiter is free from the tank

	gear_status = (gear_status == DOOR_CLOSED || gear_status == DOOR_CLOSING ?
				   DOOR_OPENING : DOOR_CLOSING);
	OperateLandingGear();
}

void Atlantis::ActivateCargoDoors (DoorStatus action)
{
	door_status = action;
}

void Atlantis::RevertCargoDoors (void)
{
	ActivateCargoDoors (door_status == DOOR_CLOSED || door_status == DOOR_CLOSING ?
		DOOR_OPENING : DOOR_CLOSING);
}

void Atlantis::OperateArm (void)
{
	if (!arm_moving) 
	{
		arm_moving = true;
		RegisterAnimation();
	}
}

void Atlantis::RevertArm (void)
{
	varm_proc=0.0;
	arm_proc =0.0;
	OperateArm();
}

void Atlantis::RotateArm ()
{
	MESHGROUP_TRANSFORM mt;
	mt.transform = MESHGROUP_TRANSFORM::ROTATE;
	mt.nmesh = 0;

	mt.P.rotparam.angle = (float) arm_wrist_roll;
	mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN;
	mt.ngrp = 53; 
	MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) arm_wrist_yaw;
	mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN;
	mt.ngrp = 53; 
	MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) arm_wrist_pitch;
	mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN;
	for (mt.ngrp = 53; mt.ngrp < 55; mt.ngrp++)
		MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) arm_elbow_pitch;
	mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN;
	for (mt.ngrp = 52; mt.ngrp < 55; mt.ngrp++)
		MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) arm_shoulder_pitch;
	mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN;
	for (mt.ngrp = 51; mt.ngrp < 56; mt.ngrp++)
		MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) arm_shoulder_yaw;
	mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
	mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN;
	for (mt.ngrp = 51; mt.ngrp < 56; mt.ngrp++)
		MeshgroupTransform (vis, mt);

	if (do_plat)
	{
		VECTOR3 plat_ofs = _V(-2.59805, 1.69209, -5.15524);
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		mt.nmesh = 1;
		if (do_cargostatic) mt.nmesh++;
		mt.ngrp = -1; // transform complete mesh

		mt.P.rotparam.angle = (float) arm_wrist_roll;
		mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_wrist_yaw;
		mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_wrist_pitch;
		mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_elbow_pitch;
		mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_shoulder_pitch;
		mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_shoulder_yaw;
		mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);
	}

	if (sat_grappled & !sat_jettisoned)
	{
		VECTOR3 plat_ofs = ofs_sts_sat;
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		mt.ngrp = -1; // transform complete mesh

		mt.nmesh = 1;
		if (do_plat)        mt.nmesh++;
		if (do_cargostatic) mt.nmesh++;

		mt.P.rotparam.angle = (float) arm_wrist_roll;
		mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_wrist_yaw;
		mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_wrist_pitch;
		mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_elbow_pitch;
		mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_shoulder_pitch;
		mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) arm_shoulder_yaw;
		mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

	}


}

void Atlantis::ZeroArm ()
{
	int grp_ofs;
	VECTOR3 plat_ofs = _V(-2.59805, 1.69209, -5.15524);
	MESHGROUP_TRANSFORM mt;
	float da =0; 

	mt.nmesh = 0;
	mt.transform = MESHGROUP_TRANSFORM::ROTATE;
	mt.P.rotparam.angle = (float) -arm_shoulder_yaw;
	mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
	mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN;
	for (grp_ofs=51;grp_ofs<56;grp_ofs++)
	{
		mt.ngrp = grp_ofs; 
		MeshgroupTransform (vis, mt);
	}
	mt.P.rotparam.angle = (float) -arm_shoulder_pitch;
	mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN;
	for (grp_ofs=51;grp_ofs<56;grp_ofs++)
	{
		mt.ngrp = grp_ofs; 
		MeshgroupTransform (vis, mt);
	}
	mt.P.rotparam.angle = (float) -arm_elbow_pitch;
	mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN;
	for (grp_ofs=52;grp_ofs<55;grp_ofs++)
	{
		mt.ngrp = grp_ofs; 
		MeshgroupTransform (vis, mt);
	}
	mt.P.rotparam.angle = (float) -arm_wrist_pitch;
	mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN;
	for (grp_ofs=53;grp_ofs<55;grp_ofs++)
	{
		mt.ngrp = grp_ofs; 
		MeshgroupTransform (vis, mt);
	}

	mt.P.rotparam.angle = (float) -arm_wrist_yaw;
	mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN;
	mt.ngrp = 53; 
	MeshgroupTransform (vis, mt);

	mt.P.rotparam.angle = (float) -arm_wrist_roll;
	mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
	mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN;
	mt.ngrp = 53; 
	MeshgroupTransform (vis, mt);

	if(do_plat)
	{
		mt.nmesh = 1;
		if (do_cargostatic) mt.nmesh++;
		mt.ngrp = -1; // transform complete mesh
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;

		mt.P.rotparam.angle = (float) -arm_shoulder_yaw;
		mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_shoulder_pitch;
		mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_elbow_pitch;
		mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_pitch;
		mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_yaw;
		mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_roll;
		mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);
	}
	if (sat_grappled & !sat_jettisoned)
	{
		VECTOR3 plat_ofs = ofs_sts_sat;
		mt.transform = MESHGROUP_TRANSFORM::ROTATE;
		mt.ngrp = -1; // transform complete mesh

		mt.nmesh = 1;
		if (do_plat)        mt.nmesh++;
		if (do_cargostatic) mt.nmesh++;

		mt.P.rotparam.angle = (float) -arm_shoulder_yaw;
		mt.P.rotparam.axis = ARM_SHOULDER_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_shoulder_pitch;
		mt.P.rotparam.axis = ARM_SHOULDER_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_SHOULDER_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_elbow_pitch;
		mt.P.rotparam.axis = ARM_ELBOW_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_ELBOW_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_pitch;
		mt.P.rotparam.axis = ARM_WRIST_PITCH_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_PITCH_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_yaw;
		mt.P.rotparam.axis = ARM_WRIST_YAW_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_YAW_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);

		mt.P.rotparam.angle = (float) -arm_wrist_roll;
		mt.P.rotparam.axis = ARM_WRIST_ROLL_VECTOR;
		mt.P.rotparam.ref  = ARM_WRIST_ROLL_ORIGIN - plat_ofs;
		MeshgroupTransform (vis, mt);
	}
}

void Atlantis::GrappleEnd (VECTOR3 *EndPoint)
{
	VECTOR3 endpoint = _V(-2.55,1.5,-5.15524);
	VECTOR3 tpt = _V(0,0,0);
	double angle;
	//VECTOR3 loc[6];
	//for (int i=0; i<6; i++)
	//	loc[i] = _V(0,0,0);
	//loc[2] = endpoint - ARM_ELBOW_PITCH_ORIGIN;
	//loc[3] = endpoint - ARM_WRIST_PITCH_ORIGIN;
	//loc[4] = endpoint - ARM_WRIST_YAW_ORIGIN;
	//loc[5] = endpoint - ARM_WRIST_ROLL_ORIGIN;
	
	tpt = endpoint - ARM_WRIST_YAW_ORIGIN;
	angle = -arm_wrist_yaw;
	endpoint.x = tpt.x * cos(angle) - tpt.z * sin(angle);
    endpoint.y = tpt.y;
	endpoint.z = tpt.x * sin(angle) + tpt.z * cos(angle);
	endpoint = endpoint + ARM_WRIST_YAW_ORIGIN;

	tpt = endpoint - ARM_WRIST_PITCH_ORIGIN;
	angle = -arm_wrist_pitch;
	endpoint.x = tpt.x;
    endpoint.y = tpt.y * cos(angle) + tpt.z * sin(angle);
	endpoint.z = tpt.z * cos(angle) - tpt.y * sin(angle);
	endpoint = endpoint + ARM_WRIST_PITCH_ORIGIN;

	tpt = endpoint - ARM_ELBOW_PITCH_ORIGIN;
	angle = -arm_elbow_pitch;
	endpoint.x = tpt.x;
    endpoint.y = tpt.y * cos(angle) + tpt.z * sin(angle);
	endpoint.z = tpt.z * cos(angle) - tpt.y * sin(angle);
	endpoint = endpoint + ARM_ELBOW_PITCH_ORIGIN;

	tpt = endpoint - ARM_SHOULDER_PITCH_ORIGIN;
	angle = -arm_shoulder_pitch;
	endpoint.x = tpt.x;
    endpoint.y = tpt.y * cos(angle) + tpt.z * sin(angle);
	endpoint.z = tpt.z * cos(angle) - tpt.y * sin(angle);
	endpoint = endpoint + ARM_SHOULDER_PITCH_ORIGIN;

	tpt = endpoint - ARM_SHOULDER_YAW_ORIGIN;
	angle = arm_shoulder_yaw;
	endpoint.x = tpt.x * cos(angle) - tpt.z * sin(angle);
    endpoint.y = tpt.y;
	endpoint.z = tpt.x * sin(angle) + tpt.z * cos(angle);
	endpoint = endpoint + ARM_SHOULDER_YAW_ORIGIN;
	
	EndPoint->x = endpoint.x;
	EndPoint->y = endpoint.y;
	EndPoint->z = endpoint.z;
}

void Atlantis::SatellitePos (VECTOR3 *pos)
{
	VECTOR3 tpt, ofs, ref = *pos;

	ofs = ARM_SHOULDER_YAW_ORIGIN - ref;
	tpt = ref - ARM_SHOULDER_YAW_ORIGIN;
	pos->x = ofs.x + tpt.x * cos(-sat_shoulder_yaw) - tpt.z * sin(-sat_shoulder_yaw);
    pos->y = ofs.y + tpt.y;
	pos->z = ofs.z + tpt.x * sin(-sat_shoulder_yaw) + tpt.z * cos(-sat_shoulder_yaw);

	ofs = ARM_SHOULDER_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(-sat_shoulder_pitch) - tpt.z * sin(-sat_shoulder_pitch);
	pos->z = ofs.z + tpt.y * sin(-sat_shoulder_pitch) + tpt.z * cos(-sat_shoulder_pitch);

	ofs = ARM_ELBOW_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(-sat_elbow_pitch) - tpt.z * sin(-sat_elbow_pitch);
	pos->z = ofs.z + tpt.y * sin(-sat_elbow_pitch) + tpt.z * cos(-sat_elbow_pitch);

	ofs = ARM_WRIST_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(-sat_wrist_pitch) - tpt.z * sin(-sat_wrist_pitch);
	pos->z = ofs.z + tpt.y * sin(-sat_wrist_pitch) + tpt.z * cos(-sat_wrist_pitch);

	ofs = ARM_WRIST_YAW_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x * cos(-sat_wrist_yaw) - tpt.z * sin(-sat_wrist_yaw);
    pos->y = ofs.y + tpt.y;
	pos->z = ofs.z + tpt.x * sin(-sat_wrist_yaw) + tpt.z * cos(-sat_wrist_yaw);


	ofs = ARM_WRIST_YAW_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x * cos(arm_wrist_yaw) - tpt.z * sin(arm_wrist_yaw);
    pos->y = ofs.y + tpt.y;
	pos->z = ofs.z + tpt.x * sin(arm_wrist_yaw) + tpt.z * cos(arm_wrist_yaw);

	ofs = ARM_WRIST_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(arm_wrist_pitch) - tpt.z * sin(arm_wrist_pitch);
	pos->z = ofs.z + tpt.y * sin(arm_wrist_pitch) + tpt.z * cos(arm_wrist_pitch);

	ofs = ARM_ELBOW_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(arm_elbow_pitch) - tpt.z * sin(arm_elbow_pitch);
	pos->z = ofs.z + tpt.y * sin(arm_elbow_pitch) + tpt.z * cos(arm_elbow_pitch);

	ofs = ARM_SHOULDER_PITCH_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x;
    pos->y = ofs.y + tpt.y * cos(arm_shoulder_pitch) - tpt.z * sin(arm_shoulder_pitch);
	pos->z = ofs.z + tpt.y * sin(arm_shoulder_pitch) + tpt.z * cos(arm_shoulder_pitch);

	ofs = ARM_SHOULDER_YAW_ORIGIN - ref;
	tpt = *pos - ofs;
	pos->x = ofs.x + tpt.x * cos(arm_shoulder_yaw) - tpt.z * sin(arm_shoulder_yaw);
    pos->y = ofs.y + tpt.y;
	pos->z = ofs.z + tpt.x * sin(arm_shoulder_yaw) + tpt.z * cos(arm_shoulder_yaw);
}

void Atlantis::SatelliteRot (VECTOR3 *d)
{
	VECTOR3 tpt;

	tpt.x = d->x * cos (-sat_shoulder_yaw) - d->z * sin (-sat_shoulder_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (-sat_shoulder_yaw) + d->z * cos (-sat_shoulder_yaw);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (-sat_shoulder_pitch) - d->z * sin (-sat_shoulder_pitch);
	tpt.z = d->y * sin (-sat_shoulder_pitch) + d->z * cos (-sat_shoulder_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (-sat_elbow_pitch) - d->z * sin (-sat_elbow_pitch);
	tpt.z = d->y * sin (-sat_elbow_pitch) + d->z * cos (-sat_elbow_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (-sat_wrist_pitch) - d->z * sin (-sat_wrist_pitch);
	tpt.z = d->y * sin (-sat_wrist_pitch) + d->z * cos (-sat_wrist_pitch);
	*d = tpt;

	tpt.x = d->x * cos (-sat_wrist_yaw) - d->z * sin (-sat_wrist_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (-sat_wrist_yaw) + d->z * cos (-sat_wrist_yaw);
	*d = tpt;
	
	tpt.x = d->x * cos (arm_wrist_yaw) - d->z * sin (arm_wrist_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (arm_wrist_yaw) + d->z * cos (arm_wrist_yaw);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_wrist_pitch) - d->z * sin (arm_wrist_pitch);
	tpt.z = d->y * sin (arm_wrist_pitch) + d->z * cos (arm_wrist_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_elbow_pitch) - d->z * sin (arm_elbow_pitch);
	tpt.z = d->y * sin (arm_elbow_pitch) + d->z * cos (arm_elbow_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_shoulder_pitch) - d->z * sin (arm_shoulder_pitch);
	tpt.z = d->y * sin (arm_shoulder_pitch) + d->z * cos (arm_shoulder_pitch);
	*d = tpt;

	tpt.x = d->x * cos (arm_shoulder_yaw) - d->z * sin (arm_shoulder_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (arm_shoulder_yaw) + d->z * cos (arm_shoulder_yaw);
	*d = tpt;
}

void Atlantis::GrappleNormal (VECTOR3 *d)
{
	VECTOR3 tpt;
	*d = _V(0,0,-1);

	tpt.x = d->x * cos (arm_wrist_yaw) - d->z * sin (arm_wrist_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (arm_wrist_yaw) + d->z * cos (arm_wrist_yaw);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_wrist_pitch) - d->z * sin (arm_wrist_pitch);
	tpt.z = d->y * sin (arm_wrist_pitch) + d->z * cos (arm_wrist_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_elbow_pitch) - d->z * sin (arm_elbow_pitch);
	tpt.z = d->y * sin (arm_elbow_pitch) + d->z * cos (arm_elbow_pitch);
	*d = tpt;

	tpt.x = d->x;
	tpt.y = d->y * cos (arm_shoulder_pitch) - d->z * sin (arm_shoulder_pitch);
	tpt.z = d->y * sin (arm_shoulder_pitch) + d->z * cos (arm_shoulder_pitch);
	*d = tpt;

	tpt.x = d->x * cos (arm_shoulder_yaw) - d->z * sin (arm_shoulder_yaw);
	tpt.y = d->y;
	tpt.z = d->x * sin (arm_shoulder_yaw) + d->z * cos (arm_shoulder_yaw);
	*d = tpt;
}

// ==============================================================
// API interface
// ==============================================================

// Initialisation
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new Atlantis (hvessel, flightmodel);
}

// Cleanup
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (Atlantis*)vessel;
}

// Set the capabilities of the vessel class
DLLCLBK void ovcSetClassCaps (VESSEL *vessel, FILEHANDLE cfg)
{
	//((Atlantis*)vessel)->SetLaunchConfiguration();
}

// Read status from scenario file
DLLCLBK void ovcLoadStateEx (VESSEL *vessel, FILEHANDLE scn, void *vs)
{
	((Atlantis*)vessel)->LoadState (scn, vs);
}

// Save status to scenario file
DLLCLBK void ovcSaveState (VESSEL *vessel, FILEHANDLE scn)
{
	((Atlantis*)vessel)->SaveState (scn);
}

DLLCLBK void ovcTimestep (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->Timestep (simt);
}

DLLCLBK void ovcAnimate (VESSEL *vessel, double simt)
{
	((Atlantis*)vessel)->UpdateMesh ();
}

DLLCLBK void ovcVisualCreated (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	if (refcount > 1) return; // we don't allow more than one visual per object
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = vis;
	sts->vdoor_proc = 0.0;
	sts->vgear_proc = 0.0;
	sts->varm_proc = 0.0;
	sts->UpdateMesh ();
	sts->RotateArm();
}

DLLCLBK void ovcVisualDestroyed (VESSEL *vessel, VISHANDLE vis, int refcount)
{
	Atlantis *sts = (Atlantis*)vessel;
	sts->vis = NULL;
}

// Keyboard interface handler (buffered key events)
DLLCLBK int ovcConsumeBufferedKey (VESSEL *vessel, DWORD key, bool down, char *kstate)
{
	if (!down) return 0; // only process keydown events

	Atlantis *sts = (Atlantis*)vessel;

	if (KEYMOD_SHIFT (kstate)) {

		switch (key) {
		case OAPI_KEY_J: // "Jettison"
			sts->bManualSeparate = true;
			sts->do_switch = false;
			return 1;
		case OAPI_KEY_1: 
			if (sts->arm_shoulder_pitch - sts->arm_angle > ARM_SHOULDER_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;		
		case OAPI_KEY_2: 
			if (sts->arm_shoulder_yaw - sts->arm_angle > ARM_SHOULDER_YAW_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_YAW;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_3:
			if (sts->arm_elbow_pitch - sts->arm_angle > ARM_ELBOW_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_ELBOW_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_4:
			if (sts->arm_wrist_pitch - sts->arm_angle > ARM_WRIST_PITCH_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_PITCH;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_5:
			if (sts->arm_wrist_yaw - sts->arm_angle > ARM_WRIST_YAW_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_YAW;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_6: 
			if (sts->arm_wrist_roll - sts->arm_angle > ARM_WRIST_ROLL_MIN && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_ROLL;
				sts->arm_max = -sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_E:
			double total_rotation=0;
			total_rotation += sts->arm_shoulder_pitch;
			total_rotation += sts->arm_shoulder_yaw;
			total_rotation += sts->arm_elbow_pitch;
			total_rotation += sts->arm_wrist_pitch;
			total_rotation += sts->arm_wrist_yaw;
			total_rotation += sts->arm_wrist_roll;
			if (total_rotation ==0)
			{
				if (sts->do_plat)
				{
					sts->do_plat=false;
				}
				else
				{
					sts->do_plat=true;
				}
				sts->ClearMeshes();
				sts->ClearExhaustRefs();
				sts->ClearAttExhaustRefs();
				sts->AddOrbiterVisual (OFS_ZERO, !sts->sat_jettisoned);
			}
			return 1;
		}	
	} else if (KEYMOD_CONTROL (kstate)) {
		// insert ctrl key combinations here
	} else { // unmodified keys
		switch (key) {
		case OAPI_KEY_G:  // "Landing gear"
			sts->RevertLandingGear ();
			return 1;
		case OAPI_KEY_J:  // "Jettison"
			sts->bManualSeparate = true;
			return 1;
		case OAPI_KEY_K:  // "Cargo bay doors"
			sts->RevertCargoDoors ();
			return 1;
		case OAPI_KEY_1:
			if (sts->arm_shoulder_pitch + sts->arm_angle < ARM_SHOULDER_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_2: 
			if (sts->arm_shoulder_yaw + sts->arm_angle < ARM_SHOULDER_YAW_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_SHOULDER_YAW;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_3: 
			if (sts->arm_elbow_pitch + sts->arm_angle < ARM_ELBOW_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_ELBOW_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_4: 
			if (sts->arm_wrist_pitch + sts->arm_angle < ARM_WRIST_PITCH_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_PITCH;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_5: 
			if (sts->arm_wrist_yaw + sts->arm_angle < ARM_WRIST_YAW_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_YAW;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_6:
			if (sts->arm_wrist_roll + sts->arm_angle < ARM_WRIST_ROLL_MAX && !sts->arm_moving)
			{
				sts->arm_action = sts->ARM_WRIST_ROLL;
				sts->arm_max = sts->arm_angle;
				sts->RevertArm();
			}
			return 1;
		case OAPI_KEY_8:
			sts->ToggleGrapple();
			return 1;
		case OAPI_KEY_9: 
			sts->center_arm = true;
			return 1;
		case OAPI_KEY_0: 
			if (sts->arm_angle == ARM_ANGLE_HIGH)
				sts->arm_angle = ARM_ANGLE_MED;
			else if (sts->arm_angle == ARM_ANGLE_MED)
				sts->arm_angle = ARM_ANGLE_LOW;
			else if (sts->arm_angle == ARM_ANGLE_LOW)
				sts->arm_angle = ARM_ANGLE_HIGH;
			return 1;
		case OAPI_KEY_E: 
			sts->do_eva=true;
			return 1;
		}
	}
	return 0;
}
