// ==============================================================
//                 ORBITER MODULE: Atlantis
//                  Part of the ORBITER SDK
//          Copyright (C) 2001-2002 Martin Schweiger
//                   All rights reserved
//
// Atlantis.h
// Class interface of Atlantis (Space Shuttle) vessel class
// module and associated subclasses (SRB, tank)
// ==============================================================

#ifndef __ATLANTIS_H
#define __ATLANTIS_H

#include "orbitersdk.h"
#include <math.h>

// ==========================================================
// Some Orbiter-related parameters
// ==========================================================

const double ORBITER_EMPTY_MASS = 104326.0;
// Orbiter empty mass [kg]

const double ORBITER_MAX_PROPELLANT_MASS = 20000.0;
// Amount of fuel the orbiter can hold in internal tanks
// I don't have this figure, so this is purely speculative ...

const double ORBITER_MAIN_THRUST = 1668652.0 * 1.25;
// Vacuum thrust rating per main engine [N] (x3 for total)
// assuming vacuum thrust is 5/4 liftoff thrust

const double ORBITER_OMS_THRUST = 26700.0;
// Vacuum thrust per unit for Orbital Maneuvering System [N] (x2 for total)

const double ORBITER_RCS_THRUST = 7740.0;
// Vacuum thrust rating for attitude thrusters (Reaction Control System) [N]

const double ORBITER_MAIN_ISP0 = 5000.0;
const double ORBITER_MAIN_ISP1 = 4000.0;
// Vacuum and sea-level fuel-specific impulse for orbiter main engines [m/s]

const double ORBITER_OMS_ISP0 = 5000.0;
const double ORBITER_OMS_ISP1 = 4000.0;
// Vacuum and sea-level fuel-specific impulse for Orbital Maneuvering System [m/s]
// In fact this should be the value for monomethyl hydrazine/nitrogen tetroxide

const double ORBITER_RCS_ISP0 = 5000.0;
const double ORBITER_RCS_ISP1 = 4000.0;
// Vacuum and sea-level fuel-specific impulse for Reaction Control System [m/s]

const double GEAR_OPERATING_SPEED = 0.3;
// Opening/closing speed of landing gear (1/sec)
// => gear cycle ~ 3 sec

const double DOOR_OPERATING_SPEED = 0.00952;
// Opening/closing speed of payload bay doors (1/sec)
// This contains both the door opening sequence (63 sec) and the
// front radiator deployment sequence

const double ARM_OPERATING_SPEED = 0.05;
const double ARM_MAX_ANGLE = RAD*90.0;
// ARM PARAMETERS

const double ARM_ANGLE_HIGH = RAD*10.0;
const double ARM_ANGLE_MED  = RAD*1.0;
const double ARM_ANGLE_LOW  = RAD*0.1;
//First joint

const VECTOR3 ARM_SHOULDER_PITCH_ORIGIN = _V(-2.55,1.5,10.4186);
const VECTOR3 ARM_SHOULDER_PITCH_VECTOR = _V(1,0,0);
const double  ARM_SHOULDER_PITCH_MAX    = RAD*145.0;
const double  ARM_SHOULDER_PITCH_MIN    = RAD*-2.0;
const int     ARM_SHOULDER_PITCH_START  = 51;
const int     ARM_SHOULDER_PITCH_END    = 55;

const VECTOR3 ARM_SHOULDER_YAW_ORIGIN   = _V(-2.55,1.5,10.4186);
const VECTOR3 ARM_SHOULDER_YAW_VECTOR   = _V(0,1,0);
const double  ARM_SHOULDER_YAW_MAX      = RAD*180.0;
const double  ARM_SHOULDER_YAW_MIN      = RAD*-180.0;
const int     ARM_SHOULDER_YAW_START    = 51;
const int     ARM_SHOULDER_YAW_END      = 55;

const VECTOR3 ARM_ELBOW_PITCH_ORIGIN    = _V(-2.55,1.5,3.5893);
const VECTOR3 ARM_ELBOW_PITCH_VECTOR    = _V(1,0,0);
const double  ARM_ELBOW_PITCH_MAX       = RAD*2.0;
const double  ARM_ELBOW_PITCH_MIN       = RAD*-160.0;
const int     ARM_ELBOW_PITCH_START     = 52;
const int     ARM_ELBOW_PITCH_END       = 54;

const VECTOR3 ARM_WRIST_PITCH_ORIGIN    = _V(-2.55,1.5,-2.45965);
const VECTOR3 ARM_WRIST_PITCH_VECTOR    = _V(1,0,0);
const double  ARM_WRIST_PITCH_MAX       = RAD*120.0;
const double  ARM_WRIST_PITCH_MIN       = RAD*-120.0;
const int     ARM_WRIST_PITCH_START     = 53;
const int     ARM_WRIST_PITCH_END       = 54;

const VECTOR3 ARM_WRIST_YAW_ORIGIN      = _V(-2.55,1.5,-3.27446);
const VECTOR3 ARM_WRIST_YAW_VECTOR      = _V(0,1,0);
const double  ARM_WRIST_YAW_MAX         = RAD*120.0;
const double  ARM_WRIST_YAW_MIN         = RAD*-120.0;
const int     ARM_WRIST_YAW_START       = 53;
const int     ARM_WRIST_YAW_END         = 53;

const VECTOR3 ARM_WRIST_ROLL_ORIGIN     = _V(-2.55,1.5,-3.27446);
const VECTOR3 ARM_WRIST_ROLL_VECTOR     = _V(0,0,-1);
const double  ARM_WRIST_ROLL_MAX        = RAD*447.0;
const double  ARM_WRIST_ROLL_MIN        = RAD*-447.0;
const int     ARM_WRIST_ROLL_START      = 53;
const int     ARM_WRIST_ROLL_END        = 53;

const VECTOR3 ORBITER_CS = {234.8,389.1,68.2};
// Orbiter cross sections (projections into principal axes) [m^2]

const VECTOR3 ORBITER_CS_GEAR = {10.0,0.0,3.0};
// Contribution of fully extended landing gear to cross sections

const double ORBITER_CW[4] = {0.13, 0.5, 1.4, 1.4};
// Orbiter wind resistance coefficients in principal directions (+z, -z, +-x, +-y)

const double ORBITER_CW_GEAR[4] = {0.04, 0.04, 0.05, 0.0};
// Contribution of fully extended landing gear to wind resistance

// ==========================================================
// Some Tank-related parameters
// ==========================================================

const double TANK_MAX_PROPELLANT_MASS = 719115.0;
// Main tank propellant mass [kg]

const double TANK_EMPTY_MASS = 35425.0;
// Main tank empty mass

// ==========================================================
// Some SRB-related parameters
// ==========================================================

const double SRB_MAX_PROPELLANT_MASS = 502126.0;
// SRB propellant mass [kg]

const double SRB_EMPTY_MASS = 87543.0;
// SRB empty mass [kg]

const double SRB_ISP0 = 3574.68;
const double SRB_ISP1 = 2859.74;
// SRB vacuum and sea-level fuel-specific impulse [m/s]

const double SRB_THRUST = 1202020.0*9.81 * 1.25;
// Vacuum SRB thrust per unit [N]

const double MAX_ATT_LAUNCH = 1e5;
// Max attitude thrust during launch phase (temporary)

const double SRB_STABILISATION_TIME = 4.0;
// MET: -SRB ignition

const double SRB_SEPARATION_TIME = 126.0;
// MET: SRB separation

const double SRB_CUTOUT_TIME = 135.0;
// MET: engine shutdown

// ==========================================================
// Mesh offsets for various configurations
// ==========================================================

const VECTOR3 OFS_ZERO             = { 0.0, 0.0,  0.0  };
const VECTOR3 OFS_LAUNCH_ORBITER   = { 0.0, 6.22,-7.795};
const VECTOR3 OFS_LAUNCH_TANK      = { 0.0,-1.91, 5.72 };
const VECTOR3 OFS_LAUNCH_RIGHTSRB  = { 6.8,-1.91,-5.68 };
const VECTOR3 OFS_LAUNCH_LEFTSRB   = {-6.8,-1.91,-5.68 };
const VECTOR3 OFS_WITHTANK_ORBITER = { 0.0, 4.79,-9.185};
const VECTOR3 OFS_WITHTANK_TANK    = { 0.0,-3.34, 4.33 };
const VECTOR3 ORBITER_DOCKPOS      = { 0.0, 2.44,10.44 };
const VECTOR3 OFS_MMU              = {0,2.44,10.44};

// ==========================================================
// Mesh group indices for some components
// ==========================================================

const int ORBITER_MESHGROUP_GEAR = 96;

// ==========================================================
// Interface for derived vessel class: Atlantis
// ==========================================================

class Atlantis: public VESSEL {
public:
	enum DoorStatus { DOOR_CLOSED, DOOR_OPEN, DOOR_CLOSING, DOOR_OPENING } door_status, gear_status;
	Atlantis (OBJHANDLE hObj, int fmodel);
	void LoadState (FILEHANDLE scn, void *vs);
	void SaveState (FILEHANDLE scn);
	void SetLaunchConfiguration (void);
	void SetPostLaunchConfiguration (double srbtime);
	void SetOrbiterTankConfiguration (void);
	void SetOrbiterConfiguration (void);
	void DefineAnimations (void);
	void AddOrbiterVisual (const VECTOR3 &ofs, bool add_sat);
	void AddTankVisual (const VECTOR3 &ofs);
	void AddSRBVisual (const VECTOR3 &ofs);
	void SeparateBoosters (double srb_time);
	void SeparateTank (void);
	void SeparateSatellite (void);
	void SeparateMMU (void);
	void ToggleGrapple (void);
	void Timestep (double simt);
	void Jettison ();
	void UpdateMesh ();
	void OperateLandingGear ();
	void RevertLandingGear ();
	void ActivateCargoDoors (DoorStatus action);
	void RevertCargoDoors ();
	void OperateArm ();
	void RevertArm ();
	void RotateArm ();
	void ZeroArm ();
	void GrappleEnd(VECTOR3 *EndPoint);
	void GrappleNormal(VECTOR3 *Normal);
	void SatellitePos (VECTOR3 *ref);
	void SatelliteRot (VECTOR3 *d);
	int status; // 0=launch configuration
	            // 1=SRB's engaged
	            // 2=SRB's separated
	            // 3=Tank separated (orbiter only)
	bool bManualSeparate; // flag for user-induced booster or tank separation

	double t0;          // reference time: designated liftoff time
	WORD srb_id1, srb_id2;

	double door_proc, vdoor_proc; // logical and visual payload bay door opening state (0=closed, 1=open)
	double gear_proc, vgear_proc;
	bool gear_moving, gear_endanim;
	double arm_proc, varm_proc, arm_max, arm_min;
	double arm_angle;
	int arm_start, arm_end;
	bool center_arm;
	VECTOR3 arm_origin;
	VECTOR3 arm_axis;
	double varm_shoulder_pitch, varm_shoulder_yaw, varm_elbow_pitch, varm_wrist_pitch, varm_wrist_yaw, varm_wrist_roll;
	double arm_shoulder_pitch, arm_shoulder_yaw, arm_elbow_pitch, arm_wrist_pitch, arm_wrist_yaw, arm_wrist_roll;
	double sat_shoulder_pitch, sat_shoulder_yaw, sat_elbow_pitch, sat_wrist_pitch, sat_wrist_yaw, sat_wrist_roll;
	bool arm_moving, arm_endanim;
	VECTOR3 sat_grapple;
	VECTOR3 sat_grapple_norm;
	VECTOR3 cargo_static_ofs;
	bool sat_grappled;
	bool sat_jettisoned;
	bool do_switch;
	bool do_eva;
	bool do_plat;
	bool do_cargostatic;
	VECTOR3 sat_ofs;
	VECTOR3 ofs_sts_sat;
	VISHANDLE vis;      // handle for visual - note: we assume that only one visual per object is created!
	MESHHANDLE hOrbiterMesh, hTankMesh, hSRBMesh; // mesh handles
	enum ArmStatus { ARM_CLOSED, ARM_OPEN, ARM_CLOSING, ARM_OPENING } arm_status;
	enum ArmAction { ARM_SHOULDER_PITCH, ARM_SHOULDER_YAW, ARM_ELBOW_PITCH, ARM_WRIST_PITCH, ARM_WRIST_YAW, ARM_WRIST_ROLL} arm_action;
	char sat_name[256];
	char sat_mesh_name[256];
	char cargo_static_mesh_name[256];

private:
	void AutoMainGimbal();
	void LaunchClamps();
	void CreateAttControls_Launch();
	void CreateAttControls_RCS();

	UINT anim_door;                            // handle for cargo door animation
	PROPELLANT_HANDLE ph_oms, ph_tank, ph_srb; // handles for propellant resources
	THRUSTER_HANDLE th_main[3];                // handles for orbiter main engines
	THRUSTER_HANDLE th_srb[2];                 // handles for SRB engines
	THGROUP_HANDLE thg_main, thg_srb;          // handles for thruster groups
};

// ==========================================================
// Interface for derived vessel class: Atlantis_SRB
// ==========================================================

class Atlantis_SRB: public VESSEL {
public:
	Atlantis_SRB (OBJHANDLE hObj);
	// Construct interface from existing object

	void SetClassSpecs (void);
	void Timestep (double simt);
	void SetRefTime (void);

private:
	double t0;                  // reference time: liftoff
	double srb_separation_time; // simulation time at which SRB separation was initiated
	bool bMainEngine;           // main engine firing?
	bool bSeparationEngine;     // separation thrusters firing?
	MESHHANDLE hSRBMesh;
	PROPELLANT_HANDLE ph_main;  // handle for propellant resource
	THRUSTER_HANDLE th_main;    // engine handle
	THRUSTER_HANDLE th_bolt;    // separation bolt
};

// ==========================================================
// Interface for derived vessel class: Atlantis_Tank
// ==========================================================

class Atlantis_Tank: public VESSEL {
public:
	Atlantis_Tank (OBJHANDLE hObj);
	// Construct interface from existing object

	void SetClassSpecs (void);
	void Timestep (double simt);

private:
	MESHHANDLE hTankMesh;
};

#endif // !__ATLANTIS_H
